@@ -1,7 +1,8 @@
 # This file makes the directory a Python package
 from .engine import Engine
+from .analytics_components import AnalyticsComponents
 from .filter_methods import FilterMethods
 from .query_methods import QueryMethods
 from .support_methods import SupportMethods
 
-__all__ = ['Engine', 'FilterMethods', 'QueryMethods', 'SupportMethods']
+__all__ = ['Engine', 'FilterMethods', 'QueryMethods', 'AnalyticsComponents', 'SupportMethods']
