@@ -2,8 +2,8 @@
 import re
 import pandas as pd
 import numpy as np
-from ..metric import Metric, MetricGroup
-from typing import Dict, List, Any, Optional, Union, Callable
+from ..metric import Metric, MetricGroup, ComputedMetric
+from typing import Dict, List, Any, Optional
 
 def add_quotes_to_brackets(expression: str) -> str:
                     return re.sub(r'\[(.*?)\]', r"['\1']", expression)
@@ -16,6 +16,11 @@ def brackets_to_backticks(expression: str) -> str:
     return re.sub(r'\[(.*?)\]', lambda m: f"`{m.group(1)}`", expression)
 
 class QueryMethods:
+    def dimension(self, dimension:str,
+        context_state_name: str = 'Default',
+        query_filters: Optional[Dict[str, Any]] = None) -> List[str]:
+        df = self.dimensions([dimension], context_state_name=context_state_name, query_filters=query_filters)
+        return df[dimension]
 
     def dimensions(
         self,
@@ -27,9 +32,9 @@ class QueryMethods:
         if query_filters is None:
             query_filters = {}
         df = self.context_states[context_state_name].copy()
-        df = self.fetch_and_merge_columns(columns_to_fetch, df)
+        df = self._fetch_and_merge_columns(columns_to_fetch, df)
 
-        df = self.apply_filters_to_dataframe(df, query_filters)
+        df = self._apply_filters_to_dataframe(df, query_filters)
         
         if retrieve_keys:
             return df.drop_duplicates()
@@ -40,63 +45,74 @@ class QueryMethods:
         """Register variables/functions available to DataFrame.query via @name."""
         self.registered_functions.update(kwargs)
 
-    def define_metric(
-        self,
-        name: Optional[str] = None,
-        expression: Optional[str] = None,
-        aggregation: Optional[Union[str, Callable[[Any], Any]]] = None,
-        metric_filters: Optional[Dict[str, Any]] = None,
-        row_condition_expression: Optional[str] = None, 
-        context_state_name: str = 'Default',
-        fillna: Optional[any] = None, ):
-        
-        new_metric = Metric(name,expression, aggregation, metric_filters, row_condition_expression, context_state_name, fillna)
-        #if new_metric.name in self.metrics:
-            #print(f'"{new_metric.name}" is being updated')
-        self.metrics[new_metric.name] = new_metric
-
-    def define_query(
-        self,
-        query_name: str,
-        dimensions: set[str] = {},
-        metrics: List[str] = [],
-        drop_null_dimensions: bool = False,
-        drop_null_metric_results: bool = False,
-        computed_metrics: Optional[List[Dict[str, Any]]] = None,
-        having: Optional[str] = None
-    ):
-        dimensions = list(dimensions)
-        list_of_metric_objects = []
-        for metric in metrics:
-            list_of_metric_objects.append(self.metrics[metric])
-        self.queries[query_name] = {
-            "dimensions": dimensions,
-            "metrics": list_of_metric_objects,
-            "drop_null_dimensions": drop_null_dimensions,
-            "drop_null_metric_results": drop_null_metric_results,
-            "computed_metrics": computed_metrics or [],
-            "having": having
-        }
-
-    def query(self, query_name: str):
+    def query(self, query_name: str, return_internal_metrics: bool = False):
         query = self.queries.get(query_name)
         if not query:
             raise ValueError(f"Query '{query_name}' not found.")
-        # Execute the query
+        
+        # Build full set of metrics to compute:
+        # - Base metrics the query asks for
+        # - Base metrics required by computed metrics' expressions
+        # - Base metrics required by HAVING expression
+        missing_metrics: List[str] = []
+        # From computed metrics' columns
+        for cm_name in query["computed_metrics"]:
+            cm_obj = self.computed_metrics[cm_name]
+            for col in getattr(cm_obj, 'columns', []) or []:
+                if col in self.metrics and col not in query["metrics"] and col not in missing_metrics:
+                    missing_metrics.append(col)
+        # From HAVING's referenced columns (precomputed at define_query)
+        for col in query["having_columns"]:
+            if col in self.metrics and col not in query["metrics"] and col not in missing_metrics:
+                missing_metrics.append(col)
+        # From SORT's referenced columns (metrics that need to be computed for sorting)
+        for sort_col, _dir in query["sort"]:
+            if sort_col in self.metrics and sort_col not in query["metrics"] and sort_col not in missing_metrics:
+                missing_metrics.append(sort_col)
+
+        all_metrics = query["metrics"] + missing_metrics
+        metric_objects: List[Metric] = []
+        for name in all_metrics:
+            try:
+                metric_objects.append(self.metrics[name])
+            except KeyError:
+                raise ValueError(f"Metric '{name}' is not defined. Define it with define_metric().")
+        # Execute the query up until aggregation
         query_result = self._multiple_state_and_filter_query(
             dimensions = query["dimensions"],
-            metrics = query["metrics"],
+            metrics = metric_objects,
             drop_null_dimensions = query["drop_null_dimensions"],
             drop_null_metric_results = query["drop_null_metric_results"]
         )
-        # Post-aggregation computed metrics and HAVING-like filter
+        # Apply post-aggregation operations
+        # We may also need to compute computed metrics referenced by HAVING/SORT even if not requested for output
+        missing_computed_metrics: List[str] = []
+        for col in query["having_columns"]:
+            if col in self.computed_metrics and col not in query["computed_metrics"]:
+                missing_computed_metrics.append(col)
+        for sort_col, _dir in query["sort"]:
+            if sort_col in self.computed_metrics and sort_col not in query["computed_metrics"] and sort_col not in missing_computed_metrics:
+                missing_computed_metrics.append(sort_col)
+
+        computed_for_eval = query["computed_metrics"] + missing_computed_metrics
         query_result = self._apply_computed_metrics_and_having(
             query_result,
-            query.get("computed_metrics", []),
-            query.get("having")
+            computed_for_eval,
+            query["having"]
         )
-        return query_result
-    
+
+        # Sort if requested (list of tuples: [(column, 'asc'|'desc'), ...])
+        if query["sort"]:
+            by = [c for c, _ in query["sort"]]
+            ascending = [str(d).lower() == 'asc' for _, d in query["sort"]]
+            query_result = query_result.sort_values(by=by, ascending=ascending)
+
+        if return_internal_metrics:
+            return query_result
+        else:
+            # Return only requested dimensions, metrics, and computed metrics
+            return query_result[query["dimensions"] + query["metrics"] + query["computed_metrics"]]
+        
         # I use this method to avoid calling and processing the indexes multiple times when there are multiple metrics with the same context_states and filters, which I assume is the most common case.
     
     def _single_state_and_filter_query(
@@ -118,9 +134,9 @@ class QueryMethods:
                 df = self.dimensions(dimensions, True, context_state_name, query_filters)          
             else:
                 df = self.context_states[context_state_name].copy() # I copy the state DataFrame to avoid modifying the original state
-                df = self.apply_filters_to_dataframe(df, query_filters)
-                df['all'] = 1
-                dimensions = ['all']
+                df = self._apply_filters_to_dataframe(df, query_filters)
+                df['<all>'] = 1
+                dimensions = ['<all>']
                 no_dimension = True
             results = []          
             for metric in metrics:
@@ -138,7 +154,7 @@ class QueryMethods:
                     
                 metric_tables_dict = {table_name: self.tables[table_name] for table_name in metric_tables if table_name is not None}
 
-                trajectory_tables = self.find_complete_trajectory(metric_tables_dict)
+                trajectory_tables = self._find_complete_trajectory(metric_tables_dict)
                 
                 for table_name in trajectory_tables:
                     if table_name not in self.link_tables:
@@ -154,13 +170,17 @@ class QueryMethods:
                     metric_result = df.copy()[keys_and_dimensions].drop_duplicates()
 
                 if len(metric_columns_indexes) > 1:
-                    metric_result = self.fetch_and_merge_columns(list(set(metric.columns + list(metric_columns_indexes))), metric_result, drop_duplicates=True)
+                    metric_result = self._fetch_and_merge_columns(list(set(metric.columns + list(metric_columns_indexes))), metric_result, drop_duplicates=True)
                 else:
-                    metric_result = self.fetch_and_merge_columns(metric.columns, metric_result)
+                    metric_result = self._fetch_and_merge_columns(metric.columns, metric_result)
               
+                # Store masks for NaN values before filling
+                filled_masks = {}
                 if metric.fillna is not None:
                     for col in metric.columns:
-                        metric_result[col] = metric_result[col].fillna(metric.fillna)
+                        filled_masks[col] = metric_result[col].isna()
+                        metric_result.loc[filled_masks[col], col] = metric.fillna
+
                 try:
                     # If metric has row condition filter down the data based on it
                     if metric.row_condition_expression:
@@ -173,11 +193,41 @@ class QueryMethods:
                     eval_locals = {'metric_result': metric_result}
                     eval_globals = self.registered_functions
                     metric_result[metric.name] = eval(expr, eval_globals, eval_locals)
+                    
                 except Exception as e:
                     print(f"Error evaluating metric expression: {e}")
                     return None
+                
+                # Restore original NaN values in source columns (can be relevant if a metric column is also a dimension)
+                if metric.fillna is not None:
+                    for col, mask in filled_masks.items():
+                        metric_result.loc[mask, col] = pd.NA
 
-                metric_result = metric_result.groupby(dimensions,dropna = drop_null_dimensions)[metric.name].agg(metric.aggregation).reset_index()
+                # Handle ignore_dimensions (ignore all or specific dimensions)
+                if metric.ignore_dimensions:
+                    if isinstance(metric.ignore_dimensions, list):
+                        # Exclude specified dimensions (partial ignore)
+                        group_dims = [d for d in dimensions if d not in metric.ignore_dimensions]
+                        if group_dims:
+                            # Group by remaining dimensions
+                            agg_result = metric_result.groupby(group_dims, dropna=drop_null_dimensions)[metric.name].agg(metric.aggregation).reset_index()
+                            # Get all dimension combinations to join back
+                            all_dims = df[dimensions].drop_duplicates()
+                            metric_result = pd.merge(all_dims, agg_result, on=group_dims, how='left')
+                        else:
+                            # If all dimensions are excluded, calculate grand total
+                            total_agg_val = metric_result[metric.name].agg(metric.aggregation)
+                            metric_result = df[dimensions].drop_duplicates()
+                            metric_result[metric.name] = total_agg_val
+                    else:
+                        # Boolean True - ignore all dimensions (grand total)
+                        total_agg_val = metric_result[metric.name].agg(metric.aggregation)
+                        metric_result = df[dimensions].drop_duplicates()
+                        metric_result[metric.name] = total_agg_val
+                else:
+                    # Normal aggregation with all dimensions
+                    metric_result = metric_result.groupby(dimensions, dropna=drop_null_dimensions)[metric.name].agg(metric.aggregation).reset_index()
+                
                 if drop_null_metric_results:
                     metric_result = metric_result.dropna(subset=[metric.name])
                 results.append(metric_result)
@@ -186,7 +236,7 @@ class QueryMethods:
             for result in results[1:]:
                 final_result = pd.merge(final_result, result, on=dimensions, how='outer')
             if no_dimension:
-                final_result.drop('all', axis=1, inplace=True)
+                final_result.drop('<all>', axis=1, inplace=True)
 
             return final_result
 
@@ -281,23 +331,33 @@ class QueryMethods:
     def _apply_computed_metrics_and_having(
         self,
         df: pd.DataFrame,
-        computed_metrics: Optional[List[Dict[str, Any]]] = None,
+        computed_metrics: Optional[List[str]] = None,
         having: Optional[str] = None
     ) -> pd.DataFrame:
         """
         Post-aggregation stage:
-        - computed_metrics: list of dicts [{'name': str, 'expression': str, 'fillna': Any?}]
+        - computed_metrics: list of names referencing persisted computed metrics
         - having: string expression to filter aggregated rows, also using [Column] syntax.
         """
         if df is None or df.empty:
             return df
 
-        # Evaluate computed metrics
-        for cm in (computed_metrics or []):
-            if not isinstance(cm, dict) or "name" not in cm or "expression" not in cm:
-                raise ValueError("Each computed metric must be a dict with 'name' and 'expression'.")
-            name = cm["name"]
-            expression = cm["expression"]
+        # Evaluate computed metrics (resolve names to ComputedMetric instances)
+        for name in (computed_metrics or []):
+            if name not in self.computed_metrics:
+                raise ValueError(
+                    f"Computed metric '{name}' is not defined. Define it first with define_computed_metric()."
+                )
+            cm = self.computed_metrics[name]
+            expression = cm.expression
+            fillna_value = cm.fillna
+
+            # If fill na, temporarily fill the dataframe column na values
+            filled_masks = {}
+            if fillna_value is not None:
+                for col in cm.columns:
+                    filled_masks[col] = df[col].isna()
+                    df.loc[filled_masks[col], col] = fillna_value
 
             try:
                 # Turn [col] into df['col'] for Python eval
@@ -309,11 +369,14 @@ class QueryMethods:
                 eval_globals = self.registered_functions
                 df[name] = eval(expr, eval_globals, eval_locals)
 
-                # Optional fillna for computed metric
-                if "fillna" in cm:
-                    df[name] = df[name].fillna(cm["fillna"])
             except Exception as e:
                 raise ValueError(f"Error evaluating computed metric '{name}': {e}") from e
+            
+            # Restore original NaN values
+            if fillna_value is not None:
+                for col in cm.columns:
+                    # Use the same masks to restore values
+                    df.loc[filled_masks[col], col] = pd.NA
 
         # Apply HAVING-like filter
         if having:
